# Generated by Django 3.2.7 on 2022-02-28 19:16

from django.conf import settings
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='PaymentGateway',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('thumbnail', models.ImageField(blank=True, null=True, upload_to='', verbose_name='Icon')),
                ('name', models.CharField(max_length=128, verbose_name='Name')),
                ('rate_method', models.CharField(choices=[('PERCENT', 'Percentage'), ('NOMINAL', 'Nominal')], default='PERCENT', max_length=128, verbose_name='Rate method')),
                ('transfer_fee', models.DecimalField(decimal_places=2, default=0, max_digits=10, validators=[django.core.validators.MinValueValidator(0)], verbose_name='Transfer fee')),
                ('auto_confirm', models.BooleanField(default=False, help_text='Customer should inform payment manual if false', verbose_name='Auto confirmation')),
                ('active', models.BooleanField(default=True)),
                ('polymorphic_ctype', models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='polymorphic_simpel_payments.paymentgateway_set+', to='contenttypes.contenttype')),
            ],
            options={
                'verbose_name': 'Gateway',
                'verbose_name_plural': 'Gateways',
                'db_table': 'simpel_payment_gateway',
            },
        ),
        migrations.CreateModel(
            name='CashGateway',
            fields=[
                ('paymentgateway_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='simpel_payments.paymentgateway')),
            ],
            options={
                'verbose_name': 'Cash Payment',
                'verbose_name_plural': 'Cash Payment',
                'db_table': 'simpel_payment_gateway_cash',
            },
            bases=('simpel_payments.paymentgateway',),
        ),
        migrations.CreateModel(
            name='ManualTransferGateway',
            fields=[
                ('paymentgateway_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='simpel_payments.paymentgateway')),
                ('bank_name', models.CharField(max_length=128, verbose_name='Bank Name')),
                ('bank_branch_office', models.CharField(max_length=128, verbose_name='Branch Office')),
                ('bank_account', models.CharField(max_length=128, verbose_name='Account number')),
                ('bank_holder_name', models.CharField(max_length=128, verbose_name='Account holder')),
            ],
            options={
                'verbose_name': 'Manual Transfer',
                'verbose_name_plural': 'Manual Transfers',
                'db_table': 'simpel_payment_gateway_manual_transfer',
            },
            bases=('simpel_payments.paymentgateway',),
        ),
        migrations.CreateModel(
            name='Payment',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('reg_number', models.PositiveIntegerField(blank=True, editable=False, null=True, verbose_name='Reg number')),
                ('inner_id', models.CharField(blank=True, editable=False, max_length=50, null=True, unique=True, verbose_name='Inner ID')),
                ('created_at', models.DateTimeField(default=django.utils.timezone.now, editable=False, verbose_name='created at')),
                ('date_pending', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='date pending')),
                ('date_validated', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='date validated')),
                ('date_canceled', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='date canceled')),
                ('date_approved', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='date approved')),
                ('date_rejected', models.DateTimeField(blank=True, editable=False, null=True, verbose_name='date rejected')),
                ('status', models.IntegerField(choices=[(23, 'Pending'), (17, 'Validated'), (19, 'Rejected'), (18, 'Approved'), (40, 'Canceled')], default=23, verbose_name='Status')),
                ('type', models.CharField(choices=[('RCP', 'Receipt'), ('CWD', 'Withdraw')], default='RCP', max_length=15, verbose_name='Type')),
                ('reference_id', models.CharField(max_length=255, null=True, verbose_name='Reference')),
                ('amount', models.DecimalField(decimal_places=2, default=0, max_digits=15, verbose_name='Amount')),
                ('note', models.CharField(blank=True, default='', max_length=255, null=True, verbose_name='Memo')),
                ('gateway', models.ForeignKey(blank=True, help_text='Payment gateway', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='payments', to='simpel_payments.paymentgateway')),
                ('polymorphic_ctype', models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='polymorphic_simpel_payments.payment_set+', to='contenttypes.contenttype')),
                ('reference_type', models.ForeignKey(blank=True, help_text='Please select reference type, then provide valid inner id as reference, otherwise leave it blank.', limit_choices_to={'model__in': ['invoice', 'salesorder']}, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='payments', to='contenttypes.contenttype')),
                ('user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='payments', to=settings.AUTH_USER_MODEL, verbose_name='Created By')),
            ],
            options={
                'verbose_name': 'Payment',
                'verbose_name_plural': 'Payments',
                'db_table': 'simpel_payment',
                'permissions': (('validate_payment', 'Can validate Payment'), ('approve_payment', 'Can approve Payment'), ('reject_payment', 'Can reject Payment'), ('cancel_payment', 'Can cancel Payment'), ('export_payment', 'Can export Payment'), ('import_payment', 'Can import Payment')),
            },
        ),
    ]



# Generated by Django 3.2.12 on 2022-03-28 16:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('simpel_payments', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='payment',
            name='status',
            field=models.IntegerField(choices=[(1, 'Pending'), (0, 'Validated'), (2, 'Rejected'), (3, 'Approved'), (4, 'Canceled')], default=1, verbose_name='Status'),
        ),
    ]



from django.utils.translation import gettext_lazy as _


class PaymentStatus:
    WAITING = 'waiting'
    PREAUTH = 'preauth'
    CONFIRMED = 'confirmed'
    REJECTED = 'rejected'
    REFUNDED = 'refunded'
    ERROR = 'error'
    INPUT = 'input'

    CHOICES = [
        (WAITING, _('Waiting for confirmation')),
        (PREAUTH, _('Pre-authorized')),
        (CONFIRMED, _('Confirmed')),
        (REJECTED, _('Rejected')),
        (REFUNDED, _('Refunded')),
        (ERROR, _('Error')),
        (INPUT, _('Input'))]



from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver

from simpel_payments.models import Payment


@receiver(post_save, sender=Payment)
def after_save_receipt(sender, instance, **kwargs):
    if instance.reference is not None:
        instance.reference.save()

    if instance.qrcodes.first() is None:
        instance.qrcodes.create(name="public_url", qr_data=instance.admin_url)


@receiver(post_delete, sender=Payment)
def after_delete_receipt(sender, instance, **kwargs):
    if instance.reference is not None:
        instance.reference.save()



from functools import cached_property

from actstream import action
from django.contrib.auth import get_user_model
from django.contrib.contenttypes.fields import GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.core.validators import MinValueValidator
from django.db import models, transaction
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import gettext_lazy as _  # NOQA
from polymorphic.models import PolymorphicModel
from simpel_actions import mixins
from simpel_numerators.models import NumeratorMixin, NumeratorReset
from simpel_qrcodes.models import LinkedQRCode
from simpel_utils.models.fields import CustomGenericForeignKey

from . import const

LEN_SHORT = 128
LEN_LONG = 255


class PaymentGateway(PolymorphicModel):
    class Meta:
        db_table = "simpel_payment_gateway"
        verbose_name = _("Gateway")
        verbose_name_plural = _("Gateways")

    PERCENT = "PERCENT"
    NOMINAL = "NOMINAL"

    RATE_METHOD = (
        (PERCENT, _("Percentage")),
        (NOMINAL, _("Nominal")),
    )

    thumbnail = models.ImageField(
        verbose_name="Icon",
        null=True,
        blank=True,
    )
    name = models.CharField(
        max_length=LEN_SHORT,
        verbose_name=_("Name"),
    )
    rate_method = models.CharField(
        max_length=LEN_SHORT,
        default=PERCENT,
        choices=RATE_METHOD,
        verbose_name=_("Rate method"),
    )
    transfer_fee = models.DecimalField(
        max_digits=10,
        decimal_places=2,
        default=0,
        validators=[MinValueValidator(0)],
        verbose_name=_("Transfer fee"),
    )
    auto_confirm = models.BooleanField(
        default=False,
        verbose_name=_("Auto confirmation"),
        help_text=_("Customer should inform payment manual if false"),
    )
    active = models.BooleanField(default=True)

    @cached_property
    def opts(self):
        return self.get_real_instance_class()._meta

    @cached_property
    def specific(self):
        return self.get_real_instance()

    def receive_payment(self):
        """Receive payment will create Payment"""
        raise NotImplementedError("%s must implement make_payment method")

    def clean(self):
        if self.rate_method == PaymentGateway.PERCENT and self.transfer_fee > 100:
            raise ValidationError({"transfer_fee": _("%s is not valid percent number") % self.transfer_fee})

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)

    def __str__(self):
        return "%s" % (self.specific.__str__())


class CashGateway(PaymentGateway):
    class Meta:
        db_table = "simpel_payment_gateway_cash"
        verbose_name = _("Cash Payment")
        verbose_name_plural = _("Cash Payment")

    def __str__(self):
        return "%s" % (self.name)


class ManualTransferGateway(PaymentGateway):
    class Meta:
        db_table = "simpel_payment_gateway_manual_transfer"
        verbose_name = _("Manual Transfer")
        verbose_name_plural = _("Manual Transfers")

    bank_name = models.CharField(
        max_length=LEN_SHORT,
        verbose_name=_("Bank Name"),
    )
    bank_branch_office = models.CharField(
        max_length=LEN_SHORT,
        verbose_name=_("Branch Office"),
    )
    bank_account = models.CharField(
        max_length=LEN_SHORT,
        verbose_name=_("Account number"),
    )
    bank_holder_name = models.CharField(
        max_length=LEN_SHORT,
        verbose_name=_("Account holder"),
    )

    def __str__(self):
        return "%s (%s)" % (self.name, self.bank_holder_name)


class PaymentActionMixin(
    mixins.PendingMixin,
    mixins.ValidateMixin,
    mixins.CancelMixin,
    mixins.ApproveMixin,
    mixins.RejectMixin,
):
    VALID = const.VALID
    PENDING = const.PENDING
    REJECTED = const.REJECTED
    APPROVED = const.APPROVED
    CANCELED = const.CANCELED
    STATUS_CHOICES = [
        (PENDING, _("Pending")),
        (VALID, _("Validated")),
        (REJECTED, _("Rejected")),
        (APPROVED, _("Approved")),
        (CANCELED, _("Canceled")),
    ]
    status = models.IntegerField(
        choices=STATUS_CHOICES,
        default=const.PENDING,
        verbose_name=_("Status"),
    )
    qrcodes = GenericRelation(
        LinkedQRCode,
        content_type_field="linked_object_type",
        object_id_field="linked_object_id",
    )

    class Meta:
        abstract = True

    @cached_property
    def qrcode(self):
        return self.qrcodes.first()

    @cached_property
    def validate_ignore_condition(self):
        return self.is_valid or self.is_approved

    @cached_property
    def validate_valid_condition(self):
        return self.is_pending or self.is_rejected

    @cached_property
    def approve_ignore_condition(self):
        return self.is_approved

    @cached_property
    def approve_valid_condition(self):
        return self.is_valid or self.is_canceled

    @transaction.atomic
    def approve(self, request=None):
        """Approve valid order"""
        from simpel_actions import signals

        if self.approve_ignore_condition:
            return
        if self.approve_valid_condition:
            signals.pre_approve.send(sender=self.__class__, instance=self, actor=request.user, request=request)
            self.status = self.APPROVED
            self.date_approved = timezone.now()
            self.save()
            signals.post_approve.send(sender=self.__class__, instance=self, actor=request.user, request=request)
            objects = {"action_object": self}
            if self.reference is not None:
                objects["target"] = self.reference
            action.send(request.user, verb="approve %s" % self._meta.verbose_name, **objects)
        else:
            raise PermissionError(self.get_error_msg("approved"))

    @cached_property
    def reject_ignore_condition(self):
        return self.is_rejected or self.is_approved

    @cached_property
    def reject_valid_condition(self):
        return self.is_valid or self.is_canceled

    @cached_property
    def cancel_ignore_condition(self):
        return self.is_canceled

    @cached_property
    def cancel_valid_condition(self):
        return self.is_approved


class Payment(PolymorphicModel, PaymentActionMixin, NumeratorMixin):

    RECEIPT = "RCP"
    WITHDRAW = "CWD"
    PAYMENT_TYPE_CHOICES = (
        (RECEIPT, _("Receipt")),
        (WITHDRAW, _("Withdraw")),
    )

    user = models.ForeignKey(
        get_user_model(),
        null=True,
        blank=False,
        on_delete=models.SET_NULL,
        related_name="payments",
        verbose_name=_("Created By"),
    )
    type = models.CharField(
        max_length=15,
        choices=PAYMENT_TYPE_CHOICES,
        default=RECEIPT,
        verbose_name=_("Type"),
    )

    # Reference Contentype
    reference_type = models.ForeignKey(
        ContentType,
        limit_choices_to={"model__in": ["invoice", "salesorder"]},
        null=True,
        blank=True,
        related_name="payments",
        on_delete=models.SET_NULL,
        help_text=_(
            "Please select reference type, then provide valid inner id as reference, otherwise leave it blank."
        ),
    )
    reference_id = models.CharField(
        max_length=255,
        null=True,
        blank=False,
        verbose_name=_("Reference"),
    )
    reference = CustomGenericForeignKey(
        "reference_type",
        "reference_id",
        "inner_id",
    )
    gateway = models.ForeignKey(
        PaymentGateway,
        null=True,
        blank=True,
        related_name="payments",
        on_delete=models.SET_NULL,
        help_text=_("Payment gateway"),
    )
    amount = models.DecimalField(
        default=0,
        max_digits=15,
        decimal_places=2,
        verbose_name=_("Amount"),
    )
    note = models.CharField(
        default="",
        max_length=LEN_LONG,
        null=True,
        blank=True,
        verbose_name=_("Memo"),
    )

    reset_mode = NumeratorReset.MONTHLY

    class Meta:
        db_table = "simpel_payment"
        verbose_name = _("Payment")
        verbose_name_plural = _("Payments")
        permissions = (
            ("validate_payment", "Can validate Payment"),
            ("approve_payment", "Can approve Payment"),
            ("reject_payment", "Can reject Payment"),
            ("cancel_payment", "Can cancel Payment"),
            ("export_payment", "Can export Payment"),
            ("import_payment", "Can import Payment"),
        )

    def __str__(self):
        return self.inner_id

    @cached_property
    def opts(self):
        return self.get_real_instance_class()._meta

    @cached_property
    def specific(self):
        return self.get_real_instance()

    @cached_property
    def timesince(self, now=None):
        """
        Shortcut for the ``django.utils.timesince.timesince`` function of the
        current timestamp.
        """
        from django.utils.timesince import timesince as timesince_

        return timesince_(self.created_at, now)

    @cached_property
    def is_editable(self):
        return self.is_pending

    @cached_property
    def is_cancelable(self):
        return self.is_approved

    @cached_property
    def admin_url(self):
        return reverse("admin:simpel_payments_payment_inspect", args=(self.id,))

    def get_doc_prefix(self):
        return self.type

    def format_inner_id(self):
        """Inner ID final format"""
        form = [self.get_doc_prefix(), self.format_date(form="%m%y"), self.format_number()]
        inner_id = "{}.{}.{}".format(*form)
        return setattr(self, self.inner_id_field, inner_id)

    def clean(self):
        if self.reference_type is not None and self.reference is None:
            raise ValidationError({"reference_id": _("Reference not found.")})
        if self.reference:
            if self._state.adding and not self.reference.is_payable:
                raise ValidationError({"reference_id": _("Reference status is not payable.")})
            if self.amount > self.reference.payable:
                raise ValidationError(
                    {"amount": _("Amount greater than %s payable: %s" % (self.reference, self.reference.payable))}
                )

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)



from django.contrib.contenttypes.models import ContentType
from import_export.fields import Field
from import_export.resources import ModelResource
from import_export.widgets import ForeignKeyWidget

from .models import Payment, PaymentGateway

payment_fields = [
    "type",
    "user",
    "status",
    "reg_number",
    "inner_id",
    "reference_type",
    "reference_id",
    "amount",
    "note",
    "created_at",
    "date_validated",
    "date_canceled",
    "date_rejected",
    "date_approved",
]


class PaymentResource(ModelResource):
    reference_type = Field(
        attribute="reference_type",
        column_name="reference_type",
        widget=ForeignKeyWidget(ContentType, field="model"),
    )
    gateway = Field(
        attribute="gateway",
        column_name="gateway",
        widget=ForeignKeyWidget(PaymentGateway, field="model"),
    )

    class Meta:
        model = Payment
        fields = payment_fields
        export_order = payment_fields
        import_id_fields = ("inner_id",)


# class WithdrawResource(ModelResource):
#     source = Field(
#         attribute="source",
#         column_name="source",
#         widget=ForeignKeyWidget(PaymentGateway, field="name"),
#     )
#     destination = Field(
#         attribute="destination",
#         column_name="destination",
#         widget=ForeignKeyWidget(Partner, field="inner_id"),
#     )

#     class Meta:
#         model = Withdraw
#         fields = payment_fields
#         export_order = payment_fields
#         import_id_fields = ("inner_id",)



from django.contrib import admin
from django.template.loader import render_to_string
from import_export.admin import ImportExportMixin
from polymorphic.admin import PolymorphicChildModelAdmin, PolymorphicParentModelAdmin
from simpel_actions.admin import AdminActivityMixin
from simpel_admin.base import AdminPrintViewMixin, ModelAdminMixin

from simpel_payments.resources import PaymentResource

from .models import CashGateway, ManualTransferGateway, Payment, PaymentGateway


@admin.register(PaymentGateway)
class PaymentGatewayAdmin(PolymorphicParentModelAdmin):
    menu_icon = "clipboard-list"
    search_fields = ["name"]
    list_display = ["name", "rate_method", "transfer_fee"]
    child_models = [
        CashGateway,
        ManualTransferGateway,
    ]


@admin.register(ManualTransferGateway)
class ManualTransferGatewayAdmin(PolymorphicChildModelAdmin):
    pass


@admin.register(CashGateway)
class CashPaymentGatewayAdmin(PolymorphicChildModelAdmin):
    pass


@admin.register(Payment)
class PaymentAdmin(ImportExportMixin, AdminActivityMixin, AdminPrintViewMixin, ModelAdminMixin):
    menu_icon = "clipboard-list"
    readonly_fields = ["user", "status"]
    search_fields = ["inner_id", "source"]
    date_hierarchy = "created_at"
    list_display = ["inner_id", "col_detail", "status", "object_buttons"]
    list_filter = ["created_at", "status"]
    actions = [
        "validate_action",
        "reject_action",
        "approve_action",
        "cancel_action",
    ]
    resource_class = PaymentResource
    autocomplete_fields = ["gateway"]
    inspect_template = "admin/simpel_payments/payment_inspect.html"
    print_template = "admin/simpel_payments/payment_print.html"

    def col_detail(self, obj):
        context = {"object": obj}
        return render_to_string("admin/simpel_payments/payment_line.html", context=context)

    def has_change_permission(self, request, obj=None):
        default = super().has_change_permission(request, obj)
        if obj:
            return obj.is_editable and default
        return default

    def has_cancel_permission(self, request, obj=None):
        default = super().has_cancel_permission(request, obj)
        if obj:
            return obj.is_cancelable and default
        return default

    def has_approve_permission(self, request, obj=None):
        default = super().has_approve_permission(request, obj)
        if obj:
            return default
        return default

    def has_reject_permission(self, request, obj=None):
        default = super().has_reject_permission(request, obj)
        if obj:
            return default
        return default

    def has_delete_permission(self, request, obj=None):
        default = super().has_delete_permission(request, obj)
        if obj:
            return default
        return default

    def save_model(self, request, obj, form, change):
        if obj.user is None:
            obj.user = request.user
        return super().save_model(request, obj, form, change)



from django.apps import AppConfig
from django.conf import settings
from django.db.models.signals import post_migrate
from django.utils.translation import gettext_lazy as _


class SimpelPaymentsConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = "simpel_payments"
    label = "simpel_payments"
    verbose_name = _("Payments")
    icon = "credit-card-check-outline"

    def ready(self):
        from actstream import registry

        from simpel_payments import signals  # NOQA


        registry.register(self.get_model("Payment"))
        return super().ready()


def init_app(sender, **kwargs):
    pass
